/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.jtests.clients.transaction;

import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import junit.framework.Test;
import junit.framework.TestSuite;
import org.objectweb.jonas.jtests.beans.annuaire.Personne;
import org.objectweb.jonas.jtests.beans.annuaire.PersonneHome;
import org.objectweb.jonas.jtests.beans.transacted.Simple;
import org.objectweb.jonas.jtests.beans.transacted.SimpleEHome;
import org.objectweb.jonas.jtests.beans.transacted.SimpleSHome;
import org.objectweb.jonas.jtests.util.JTestCase;

/**
 * Test for Bean Managed Transaction entity beans
 * Tests also with 2 ejbjars (different beans involved)
 * @author Philippe Durieux
 */
public class F_EntityCMT extends JTestCase {

    protected static SimpleSHome slhome = null;
    protected static SimpleEHome ec2home = null;
    protected static PersonneHome ehome = null;
    String mynum = "63801";
    String badnum = "99999";
    String newnum = "12341";
    String myname = "Phil";

    public F_EntityCMT(String name) {
        super(name);
    }

    protected void setUp() {
        super.setUp();
        String BEAN_HOME;
        if (ec2home == null || slhome == null) {
            useBeans("transacted", true);
            BEAN_HOME = "transactedSimpleEC2Home";
            try {
                ec2home = (SimpleEHome) PortableRemoteObject.narrow(ictx.lookup(BEAN_HOME), SimpleEHome.class);
            } catch (NamingException e) {
                fail("Cannot get " + BEAN_HOME + ":" + e);
            }
            BEAN_HOME = "transactedSimpleSLHome";
            try {
                slhome = (SimpleSHome) PortableRemoteObject.narrow(ictx.lookup(BEAN_HOME), SimpleSHome.class);
            } catch (NamingException e) {
                fail("Cannot get " + BEAN_HOME + ":" + e);
            }
        }
        if (ehome == null) {
            useBeans("annuaire", true);
            BEAN_HOME = "annuairePersonneECHome";
            try {
                ehome = (PersonneHome) PortableRemoteObject.narrow(ictx.lookup(BEAN_HOME), PersonneHome.class);
            } catch (NamingException e) {
                fail("Cannot get " + BEAN_HOME + ":" + e);
            }
        }
    }

    public void testRollbackWithCommitInsideSL() throws Exception {
        Personne acc = ehome.create(myname, mynum);
        Simple s1 = slhome.create();
        Simple s2 = ec2home.create(300);
        try {
            utx.begin();
            acc.setNumero(badnum);
            // call another bean with new Tx committed.
            assertEquals(true, s1.opwith_requires_new());
            utx.rollback();
            assertEquals(mynum, acc.getNumero());
        } finally {
            // cleaning
            acc.remove();
            s1.remove();
            s2.remove();
        }
    }

    public void testRollbackWithCommitInsideEC2() throws Exception {
        Personne acc = ehome.create(myname, mynum);
        Simple s1 = slhome.create();
        Simple s2 = ec2home.create(300);
        try {
            utx.begin();
            acc.setNumero(badnum);
            // call another bean with new Tx committed.
            assertEquals(true, s2.opwith_requires_new());
            utx.rollback();
            assertEquals(mynum, acc.getNumero());
        } finally {
            // cleaning
            acc.remove();
            s1.remove();
            s2.remove();
        }
    }

    public void testRollbackWithCommitInsideSLEC2() throws Exception {
        Personne acc = ehome.create(myname, mynum);
        Simple s1 = slhome.create();
        Simple s2 = ec2home.create(300);
        try {
            utx.begin();
            acc.setNumero(badnum);
            // call another bean with new Tx committed.
            assertEquals(true, s1.call_requires_new_on(s2));
            utx.rollback();
            assertEquals(mynum, acc.getNumero());
        } finally {
            // cleaning
            acc.remove();
            s1.remove();
            s2.remove();
        }
    }

    /**
     * Test that rollback will not lost old bean state
     * - create bean
     * - make a transaction that change the num and roll it back
     * - check num has not been changed.
     */
    public void testStateAfterRollBack() throws Exception {
        Personne acc = ehome.create(myname, mynum);
        utx.begin();
        acc.setNumero(badnum);
        utx.rollback();
        assertEquals(mynum, acc.getNumero());
        acc.remove();		// cleaning
    }

    /**
     * Test that rollback will not lost old bean state
     * - access bean 
     * - make a transaction rolled back
     * - check state has not been changed.
     */
    public void testStateAfterRollBackNoTx() throws Exception {
        Personne acc = ehome.create(myname, mynum);
        acc.setNumero(newnum);
        utx.begin();
        acc.setNumero(badnum);
        utx.rollback();
        assertEquals(newnum, acc.getNumero());
        acc.remove();		// cleaning
    }

    /**
     * Test that rollback will not lost old bean state
     * - access bean 
     * - make a transaction rolled back
     * - check state has not been changed.
     */
    public void testStateAfterRollBackTx() throws Exception {
        Personne acc = ehome.create(myname, mynum);
        utx.begin();
        acc.setNumero(newnum);
        utx.commit();
        utx.begin();
        acc.setNumero(badnum);
        utx.rollback();
        assertEquals(newnum, acc.getNumero());
        acc.remove();		// cleaning
    }

    public static Test suite() {
        return new TestSuite(F_EntityCMT.class);
    }

    public static void main (String args[]) {
        String testtorun = null;
        // Get args
        for (int argn = 0; argn < args.length; argn++) {
            String sarg = args[argn];
            if (sarg.equals("-n")) {
                testtorun = args[++argn];
            }
        }
        if (testtorun == null) {
            junit.textui.TestRunner.run(suite());
        } else {
            junit.textui.TestRunner.run(new F_EntityCMT(testtorun));
        }
    }
}
